open Lwt

let string_of_char c =
  String.make 1 c

type xml_event =
  | Start of Xml.qname * (Xml.qname * string) list
  | End of Xml.qname
  | CData of string

let separator = ':'

let separate_namespace s =
  try
    let i = String.rindex s separator in
    let namespace = String.sub s 0 i in
    let name = String.sub s (i+1) (String.length s - (i+1)) in
    namespace,name
  with
    | Not_found -> "",s

let parse_stream s =
  let xml_parser = Expat.parser_create_ns ~separator ~encoding:None in
  Expat.parse xml_parser "<not_document>"; (* to give a root node *)
  let l = ref [] in
  let char_list = ref [] in
  let add v = match !char_list with
    | [] -> l:=v::!l
    | s -> l:=v::(CData (String.concat "" ( List.rev s )))::!l; char_list := []
  in
  let start_f s lst =
    let f (attr,value) = (separate_namespace attr,value) in
    add (Start (separate_namespace s,List.map f lst)) in
  let end_f s = add (End (separate_namespace s)) in
  let cdata_f s = char_list:=s::!char_list in
  Expat.set_start_element_handler xml_parser start_f;
  Expat.set_end_element_handler xml_parser end_f;
  Expat.set_character_data_handler xml_parser cdata_f;
  let f str =
    Expat.parse xml_parser str;
    let l' = !l in
    l := [];
    List.rev (l')
  in
  (* ce serait plus élégant un fold *)
  Lwt_stream.map_list f s

let parse_input input = 
  let xml_parser = Expat.parser_create_ns ~separator ~encoding:None in
  let q = Queue.create () in
  let char_list = ref [] in
  let add v = match !char_list with
    | [] -> Queue.push v q
    | s ->
	Queue.push (CData (String.concat "" ( List.rev s ))) q;
	Queue.push v q;
	char_list := []
  in
  let start_f s lst =
    let f (attr,value) = (separate_namespace attr,value) in
    add (Start (separate_namespace s,List.map f lst)) in
  let end_f s = add (End (separate_namespace s)) in
  let cdata_f s = char_list:=s::!char_list in
  Expat.set_start_element_handler xml_parser start_f;
  Expat.set_end_element_handler xml_parser end_f;
  Expat.set_character_data_handler xml_parser cdata_f;

  let rec f () =
    let s = String.create 256 in
    try
      let v = Queue.pop q in
      return (Some v)
    with
      | Queue.Empty ->
	  (input s 0 256)
	  >>= ( fun length ->
		  ( try
		      Expat.parse_sub xml_parser s 0 length;
		      return ()
		    with (* if there is a parse error, it is impossible to read the end of the stream
			    the exception should be added to the queue *)
		      | exn -> fail exn )
		  >>= f )
  in
  Lwt_stream.from f

let parse_fd fd = parse_input (Lwt_unix.read fd)

let solidify l =
  let rec aux acc = function
    | [] -> acc,[]
    | Start ( name, attr )::l ->
	let sons,rest = aux [] l in
	aux ((Xml.node ~attr ~sons name) :: acc) rest
    | End s::l -> acc,l
    | CData s::l -> aux ((Xml.cdata s)::acc) l
  in
  fst (aux [] l)

let map_fold f s init =
  let accum = ref init in
  let g v =
    let res,acc = f !accum v in
    accum := acc;
    res
  in
  Lwt_stream.filter_map g s

let group_stream s =
  let f (acc,stack) event =
    match event,stack with
      | Start (name,_) as v,_ ->
	  None, ( v::acc, name::stack )
      | End (name) as v, s_name::stack when name = s_name ->
	  let acc = v::acc in
	  ( match stack with
	      | [] -> Some (List.rev acc),( [], stack )
	      | t::q -> None, ( acc, stack ) )
      | v,_ ->
	  None, ( v::acc, stack )
  in
  map_fold f s ( [], [] )

let xml_stream s =
  Lwt_stream.map_list solidify (group_stream s)
