
type namespace = string
type qname = (namespace*string) (** namespace, name *)

module QnameMap =
struct
  include Map.Make ( struct type t = qname let compare = compare end )

  let get x t =
    try Some (find x t)
    with
      | Not_found -> None

end

module QnameSet = Set.Make ( struct type t = qname let compare = compare end )

module NamespaceMap = Map.Make ( String )

type ns_map =
    {
      map : string NamespaceMap.t;
      default : string;
    }

type xml =
  | Node of qname * node_content
  | CData of string

and node_content =
    {
      attr : string QnameMap.t; (** attributes *)
      sons : xml list;
    }

let empty_content =
  {
    attr = QnameMap.empty;
    sons = []
  }

let list_of_attr map =
  let f k v acc =
    (k,v)::acc
  in
  QnameMap.fold f map []

let attr_of_list lst =
  let f acc (k,v) =
    QnameMap.add k v acc
  in
  List.fold_left f QnameMap.empty lst

let add_list_to_attr lst map =
  let f acc (k,v) =
    QnameMap.add k v acc
  in
  List.fold_left f map lst

let add_list_to_attr_option lst map =
  let f acc (k,v) =
    match v with
      | None -> acc
      | Some v -> QnameMap.add k v acc
  in
  List.fold_left f map lst

let node ?(attr=[]) ?(option_attr=[]) ?(attr_map=QnameMap.empty) ?(sons=[]) qname =
  Node ( qname,
	 {
	   attr = 
	     add_list_to_attr_option option_attr
	       (add_list_to_attr attr attr_map);
	   sons = sons;
	 } )

(* c'est pas top d'avoir une exception *)
let add_attr xml (qname,value) =
  match xml with
    | CData _ -> raise (Invalid_argument "add_attr: xml is a cdata")
    | Node ( node_qname, content ) ->
	Node ( node_qname, { content with attr = QnameMap.add qname value content.attr } )

(* c'est pas top d'avoir une exception *)
let del_attr xml qname =
  match xml with
    | CData _ -> raise (Invalid_argument "del_attr: xml is a cdata")
    | Node ( node_qname, content ) ->
	Node ( node_qname, {content with attr = QnameMap.remove qname content.attr} )

let fold_del_attr f init content =
  let aux qname value (acc,attr) =
    match f qname value acc with
      | None -> (acc,attr)
      | Some new_acc -> (new_acc,QnameMap.remove qname attr)
  in
  let acc, attr = QnameMap.fold aux content.attr (init,content.attr) in
  acc, { content with attr = attr }

let find_attr s content =
  try
    Some (QnameMap.find s content.attr)
  with
    | Not_found -> None

let find_attr_exn s content =
  QnameMap.find s content.attr

let fold_del_sons f init content =
  let rec aux acc acc_list = function
    | [] -> acc, List.rev acc_list
    | node::q ->
	begin
	  match f node acc with
	    | None -> aux acc (node::acc_list) q
	    | Some new_acc -> aux new_acc acc_list q
	end
  in
  let acc, sons = aux init [] content.sons in
  acc, { content with sons = sons }

let node_content = function
  | CData _ -> raise (Invalid_argument "Xml.content: xml is a cdata")
  | Node ( node_qname, content ) -> content

let node_cdata_son = function 
  | Node (_,{ sons = [CData s] } ) -> s
  | _ -> raise ( Invalid_argument "Xml.node_cdata_son" )

(* TODO ameliorer, si ce n'est pas le premier noeud *)
let rec find_node tags list =
  match tags with
    | [] -> raise ( Invalid_argument "Xml.find_node no tag provided ( empty list )" )
    | t::q ->
	let aux = function
	  | Node (qname,_) when qname = t -> true
	  | _ -> false
	in
	let node = List.find aux list in
	match q with
	  | [] -> node
	  | _ -> find_node q (node_content node).sons

let rec find_node_opt tags list =
  try
    Some (find_node tags list)
  with
    | Not_found -> None

(* TODO handle StringPrep *)
let cdata data =
  CData data

let match_attr map (key,value) =
  try
    (QnameMap.find key map) = value
  with
    | e -> false

let only_attr map key =
  if QnameMap.mem key map
  then
    let singleton_map =
      QnameMap.add key (QnameMap.find key map) QnameMap.empty
    in
    QnameMap.equal (=) singleton_map map
  else false

let only_attrs map keys =
  let elts = QnameMap.fold (fun key _ l -> key::l) map [] in
  List.sort compare elts = List.sort compare keys

let has_attrs map keys =
  let elts = QnameMap.fold (fun key _ s -> QnameSet.add key s) map QnameSet.empty in
  let keys = List.fold_left (fun s key -> QnameSet.add key s) QnameSet.empty keys in
  QnameSet.subset keys elts

let add_qualified_name ns_map b (namespace,name) =
    match namespace with
      | "" -> Buffer.add_string b name
      | _ ->
	  if namespace = ns_map.default
	  then Buffer.add_string b name
	  else 
	    try
	      let alias = NamespaceMap.find namespace ns_map.map in
	      Buffer.add_string b alias;
	      Buffer.add_char b ':';
	      Buffer.add_string b name
	    with
	      | Not_found -> failwith namespace

let add_attribute_to_buffer ns_map b qname value =
  Buffer.add_char b ' ';
  add_qualified_name ns_map b qname;
  Buffer.add_string b "=\'";
  Buffer.add_string b value;
  Buffer.add_char b '\''

let new_ns =
  let r = ref 0 in
  fun () ->
    incr r;
    "ns"^(string_of_int !r)

let xmlns_ns = "http://www.w3.org/2000/xmlns/"
let xml_ns = "http://www.w3.org/XML/1998/namespace"

let empty_ns_map =
  { 
    map = (NamespaceMap.add xmlns_ns "xmlns"
	     (NamespaceMap.add xml_ns "xml" NamespaceMap.empty));
    default = "";
  }

module NamespaceSet = Set.Make ( String )

let default_ns map ns =
  map.default = ns

let set_potential_default_ns ns_map ns attr =
  match ns with
    | "" -> attr,ns_map
	(* XXX mal, il faudrait lui reinitialiser les namespace a vide, mais j'ai sans doute oublie de
	   mettre les namespace sur certains noeuds shame on me *)
    | s ->
	if ns = ns_map.default
	then attr,ns_map
	else
	  if NamespaceMap.mem ns ns_map.map
	  then attr,ns_map
	  else (QnameMap.add ("","xmlns") ns attr,{ ns_map with default = s })

let list_attr_ns set attr =
  let f (namespace,_) _ set =
    if namespace <> "" then NamespaceSet.add namespace set
    else set
  in
  QnameMap.fold f attr set

let list_ns ns_map xml =
  let rec aux set = function
    | CData _ -> NamespaceSet.empty
    | Node((namespace,name),content) ->
	(* let ns_sons = List.fold_left aux set content.sons in
	let ns_content = list_attr_ns ns_sons content.attr in *)
	let ns_content = list_attr_ns set content.attr in
	(* ejabberd est bugge et a besoin de ca... ( quelle merde ) *)
	if ( default_ns ns_map namespace || namespace = "" )
	then ns_content
	else NamespaceSet.add namespace ns_content
  in
  aux NamespaceSet.empty xml

let add_bind_ns_attr ns_map ns_set attr =
  let bind_ns e (ns_map,attr) =
    let ns_alias = new_ns () in
    {ns_map with map = NamespaceMap.add e ns_alias ns_map.map},
    QnameMap.add (xmlns_ns,ns_alias) e attr
  in
  let f e acc =
    if NamespaceMap.mem e ns_map.map
    then acc
    else bind_ns e acc
  in
  NamespaceSet.fold f ns_set (ns_map,attr)

let print_xml ?(human_readable=false) ?(ns_map=empty_ns_map) xml =
  let rec aux ns_map i = function
    | CData s ->
	let b = (Buffer.create (String.length s)) in
	if human_readable then Buffer.add_string b (String.make i ' ');
	Buffer.add_string b s;
	if human_readable then Buffer.add_char b '\n';
	b
    | Node((namespace,_) as qname,content) as node ->
	let attr,ns_map = set_potential_default_ns ns_map namespace content.attr in (* ejabberd est bugge et a besoin de ca... *)
	let used_ns = list_ns ns_map node in
	let ns_map,attr = add_bind_ns_attr ns_map used_ns attr in
	let b = Buffer.create 0 in
	if human_readable
	then Buffer.add_string b (String.make i ' ');
	Buffer.add_char b '<';
	add_qualified_name ns_map b qname;
	QnameMap.iter (add_attribute_to_buffer ns_map b) attr;
	( match content.sons with
	    | [] ->
		Buffer.add_string b "/>";
		if human_readable
		then Buffer.add_char b '\n';
	    | _ ->
		Buffer.add_char b '>';
		if human_readable
		then Buffer.add_char b '\n';
		List.iter (fun son -> Buffer.add_buffer b (aux ns_map (i+1) son)) content.sons;
		if human_readable
		then Buffer.add_string b (String.make i ' ');
		Buffer.add_string b "</";
		add_qualified_name ns_map b qname;
		Buffer.add_char b '>';
		if human_readable
		then Buffer.add_char b '\n'; );
	b
  in
  Buffer.contents (aux ns_map 0 xml)

