
type namespace = string
type qname = (namespace*string) (** namespace, name *)

module QnameMap : 
sig
  include Map.S with type key = qname
  val get : key -> 'a t -> 'a option
end

type xml = private
  | Node of qname * node_content
  | CData of string

and node_content =
    {
      attr : string QnameMap.t; (** attributes *)
      sons : xml list;
    }

val cdata : string -> xml

val node :
  ?attr:(QnameMap.key * string) list ->
  ?option_attr:(QnameMap.key * ( string option )) list ->
  ?attr_map:string QnameMap.t ->
  ?sons:xml list ->
  qname -> xml

val empty_content : node_content

val list_of_attr : 'a QnameMap.t -> (qname * 'a) list
val attr_of_list : (qname * 'a) list -> 'a QnameMap.t

val add_attr : xml -> (qname * string) -> xml
  (** [add_attr (qname,value) xml] add the attribute [qname] with value
      [value] to the xml node [xml] if xml is a CData, add_attr raise an
      exception [Invalid_argument] *)

val del_attr : xml -> qname -> xml
  (** [del_attr qname xml] remove the attribute [qname] to the xml
      node [xml] if xml is a CData, del_attr raise an exception
      [Invalid_argument] *)

val fold_del_attr : ( qname -> string -> 'a -> 'a option ) -> 'a -> node_content -> ( 'a * node_content )
  (** [fold_del_attr f init xml] applies [f] to every attribute of
      [xml] if [f] returns [None], the attribute is kept, if [f]
      returns [Some acc], the accumulator is updated and the attribute
      is removed *)

val find_attr : qname -> node_content -> string option
val find_attr_exn : qname -> node_content -> string

val fold_del_sons : ( xml -> 'a -> 'a option ) -> 'a -> node_content -> ( 'a * node_content )
  (** [fold_del_sons f init xml] applies [f] to every son of [xml] if
      [f] returns [None], the son is kept, if [f] returns [Some acc],
      the accumulator is updated and the son is removed *)

val node_content : xml -> node_content
  (** [content xml] returns the content of [xml] if xml is a Node,
      else raise an exception [Invalid_argument] *)

val node_cdata_son : xml -> string

val find_node : qname list -> xml list -> xml

val find_node_opt : qname list -> xml list -> xml option

module NamespaceMap : Map.S with type key = string

type ns_map =
    { map : string NamespaceMap.t;
      default : string }

val empty_ns_map : ns_map

val print_xml :
  ?human_readable:bool -> 
  ?ns_map:ns_map ->
  xml -> string

val match_attr : 'a QnameMap.t -> qname * 'a -> bool

val only_attr : 'a QnameMap.t -> qname -> bool
  (** [only_attr attr qname] is true iff there is only one attribute
  defined wich is [qname] *)

val only_attrs : 'a QnameMap.t -> qname list -> bool
val has_attrs : 'a QnameMap.t -> qname list -> bool

val xml_ns : namespace
val xmlns_ns : namespace
